"""Panda3D material system implementation."""

from __future__ import annotations

from pykotor.engine.panda3d.materials.manager import (
    Panda3DMaterial,
    Panda3DMaterialManager,
)

__all__ = [
    "Panda3DMaterial",
    "Panda3DMaterialManager",
]
