"""LIP package: lip sync data, keyframes, and binary/XML I/O."""

from __future__ import annotations

from pykotor.resource.formats.lip.lip_data import (
    LIP,
    LIPShape,
    LIPKeyFrame,
)
from pykotor.resource.formats.lip.io_lip import (
    LIPBinaryReader,
    LIPBinaryWriter,
)
from pykotor.resource.formats.lip.io_lip_xml import (
    LIPXMLReader,
    LIPXMLWriter,
)
from pykotor.resource.formats.lip.lip_auto import (
    read_lip,
    write_lip,
    detect_lip,
    bytes_lip,
)

__all__ = [
    "LIP",
    "LIPBinaryReader",
    "LIPBinaryWriter",
    "LIPKeyFrame",
    "LIPShape",
    "LIPXMLReader",
    "LIPXMLWriter",
    "bytes_lip",
    "detect_lip",
    "read_lip",
    "write_lip",
]
