"""2DA package: TwoDA data, rows, and binary/CSV/JSON I/O."""

from __future__ import annotations
from pykotor.resource.formats.twoda.twoda_data import TwoDA, TwoDARow
from pykotor.resource.formats.twoda.io_twoda import (
    TwoDABinaryReader,
    TwoDABinaryWriter,
)
from pykotor.resource.formats.twoda.io_twoda_csv import (
    TwoDACSVReader,
    TwoDACSVWriter,
)
from pykotor.resource.formats.twoda.io_twoda_json import (
    TwoDAJSONReader,
    TwoDAJSONWriter,
)
from pykotor.resource.formats.twoda.twoda_auto import (
    detect_2da,
    bytes_2da,
    read_2da,
    write_2da,
)

__all__ = [
    "TwoDA",
    "TwoDABinaryReader",
    "TwoDABinaryWriter",
    "TwoDACSVReader",
    "TwoDACSVWriter",
    "TwoDAJSONReader",
    "TwoDAJSONWriter",
    "TwoDARow",
    "bytes_2da",
    "detect_2da",
    "read_2da",
    "write_2da",
]
