"""Agnostic system utilities for cross-platform compatibility."""

from __future__ import annotations

from utility.system.agnostics.stubs import askdirectory, askopenfilename

__all__ = ["askdirectory", "askopenfilename"]
