"""TXI (texture info) package: features, commands, and binary I/O for texture metadata."""

from __future__ import annotations

from pykotor.resource.formats.txi.txi_data import TXI, TXIFeatures, TXICommand
from pykotor.resource.formats.txi.txi_auto import read_txi, write_txi, bytes_txi
from pykotor.resource.formats.txi.io_txi import TXIBinaryReader, TXIBinaryWriter

__all__ = ["TXI", "TXIBinaryReader", "TXIBinaryWriter", "TXICommand", "TXIFeatures", "bytes_txi", "read_txi", "write_txi"]
