"""NCS package: bytecode data, I/O, compilers, decompiler, and optimizers."""

from __future__ import annotations
from pykotor.resource.formats.ncs.ncs_data import (
    NCS,
    NCSInstruction,
    NCSInstructionType,
    NCSInstructionQualifier,
    NCSInstructionTypeValue,
    NCSByteCode,
)
from pykotor.resource.formats.ncs.io_ncs import (
    NCSBinaryReader,
    NCSBinaryWriter,
)
from pykotor.resource.formats.ncs.ncs_auto import bytes_ncs, compile_nss, decompile_ncs, read_ncs, write_ncs
from pykotor.resource.formats.ncs.ncs_types import NCSType, NCSTypeCode
from pykotor.resource.formats.ncs.decompiler import NCSDecompiler, DecompileError

__all__ = [
    "NCS",
    "NCSBinaryReader",
    "NCSBinaryWriter",
    "NCSByteCode",
    "NCSInstruction",
    "NCSInstructionQualifier",
    "NCSInstructionType",
    "NCSInstructionTypeValue",
    "bytes_ncs",
    "compile_nss",
    "read_ncs",
    "write_ncs",
]
