"""GL models package: Boundary, Cube, Mesh, Node, MDL loading, and vertex format utilities."""

from __future__ import annotations

from pykotor.gl.models.boundary import Boundary
from pykotor.gl.models.cube import Cube
from pykotor.gl.models.empty import Empty
from pykotor.gl.models.mesh import Mesh
from pykotor.gl.models.model import Model
from pykotor.gl.models.mdl_converter import (
    VertexFormatRequirements,
    get_node_converter_type,
    get_node_type_priority,
    should_reverse_winding_order,
)
from pykotor.gl.models.node import Node
from pykotor.gl.models.read_mdl import gl_load_mdl, gl_load_stitched_model

__all__ = [
    "Boundary",
    "Cube",
    "Empty",
    "Mesh",
    "Model",
    "Node",
    "VertexFormatRequirements",
    "get_node_converter_type",
    "get_node_type_priority",
    "should_reverse_winding_order",
    "gl_load_mdl",
    "gl_load_stitched_model",
]
