"""LTR (letter/name generator) package: Markov chain data and binary I/O."""

from __future__ import annotations

from pykotor.resource.formats.ltr.ltr_data import LTR, LTRBlock
from pykotor.resource.formats.ltr.ltr_auto import bytes_ltr, read_ltr, write_ltr
from pykotor.resource.formats.ltr.io_ltr import (
    LTRBinaryReader,
    LTRBinaryWriter,
)

__all__ = [
    "LTR",
    "LTRBinaryReader",
    "LTRBinaryWriter",
    "LTRBlock",
    "bytes_ltr",
    "read_ltr",
    "write_ltr",
]
