"""ERF (archive) package: ERF/MOD/SAV data and binary I/O (read_erf, write_erf, bytes_erf)."""

from __future__ import annotations

from pykotor.resource.formats.erf.erf_data import (
    ERF,
    ERFResource,
    ERFType,
)
from pykotor.resource.formats.erf.io_erf import (
    ERFBinaryReader,
    ERFBinaryWriter,
)
from pykotor.resource.formats.erf.erf_auto import bytes_erf, read_erf, write_erf

__all__ = [
    "ERF",
    "ERFBinaryReader",
    "ERFBinaryWriter",
    "ERFResource",
    "ERFType",
    "bytes_erf",
    "read_erf",
    "write_erf",
]
