"""Tkinter utilities for PyKotor tools.

This package provides:
- base_app: Base classes for creating GUI applications (BaseApp, ThemedApp)
- app_entry: Entry point utilities (is_frozen, setup_sys_path, etc.)
- tooltip: Tooltip widget
- updater: Update dialog and progress utilities
- rte_editor: Rich text editor utilities
- loader_dialog: Loading dialog utilities
"""
