# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'c:\GitHub\PyKotor\Libraries\PyKotor\src\utility\gui\qt\error_report_dialog.ui'
#
# Created by: PyQt5 UI code generator 5.15.11
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.
from __future__ import annotations

from typing import TYPE_CHECKING

from qtpy.QtCore import QCoreApplication, QMetaObject
from qtpy.QtWidgets import (
    QCheckBox,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QSpacerItem,
    QTextEdit,
    QVBoxLayout,
)

if TYPE_CHECKING:
    from qtpy.QtWidgets import (
        QDialog,
    )


class Ui_Dialog:
    def setupUi(self, Dialog: QDialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(600, 500)
        self.verticalLayout = QVBoxLayout(Dialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.additionalLabel = QLabel(Dialog)
        self.additionalLabel.setObjectName("additionalLabel")
        self.verticalLayout.addWidget(self.additionalLabel)
        self.additionalText = QTextEdit(Dialog)
        self.additionalText.setPlaceholderText("")
        self.additionalText.setObjectName("additionalText")
        self.verticalLayout.addWidget(self.additionalText)
        self.attachmentsLabel = QLabel(Dialog)
        self.attachmentsLabel.setObjectName("attachmentsLabel")
        self.verticalLayout.addWidget(self.attachmentsLabel)
        self.includeLogsCheckbox = QCheckBox(Dialog)
        self.includeLogsCheckbox.setObjectName("includeLogsCheckbox")
        self.verticalLayout.addWidget(self.includeLogsCheckbox)
        spacerItem = QSpacerItem(20, 4, QSizePolicy.Minimum, QSizePolicy.Fixed)
        self.verticalLayout.addItem(spacerItem)
        self.contactLabel = QLabel(Dialog)
        self.contactLabel.setWordWrap(True)
        self.contactLabel.setObjectName("contactLabel")
        self.verticalLayout.addWidget(self.contactLabel)
        self.contactInput = QLineEdit(Dialog)
        self.contactInput.setObjectName("contactInput")
        self.verticalLayout.addWidget(self.contactInput)
        self.buttonLayout = QHBoxLayout()
        self.buttonLayout.setObjectName("buttonLayout")
        spacerItem1 = QSpacerItem(40, 20, QSizePolicy.Expanding, QSizePolicy.Minimum)
        self.buttonLayout.addItem(spacerItem1)
        self.sendButton = QPushButton(Dialog)
        self.sendButton.setObjectName("sendButton")
        self.buttonLayout.addWidget(self.sendButton)
        self.verticalLayout.addLayout(self.buttonLayout)

        self.retranslateUi(Dialog)
        QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog: QDialog):
        _translate = QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Error handler"))
        self.additionalLabel.setText(_translate("Dialog", "Provide additional information (optional)"))
        self.attachmentsLabel.setText(_translate("Dialog", "Select attachments (optional)"))
        self.includeLogsCheckbox.setText(_translate("Dialog", "Log Files"))
        self.contactLabel.setText(
            _translate(
                "Dialog",
                "How can we contact you (optional, only if you want to get a response). Your report is\n"
                "       anonymous by default, so you can provide contact info like an email address here.",
            )
        )
        self.sendButton.setText(_translate("Dialog", "Send report"))
