"""Alien (Wookiee, etc.) sound ResRef set for KotOR 1 and TSL."""

from __future__ import annotations

ALIEN_SOUNDS = {  # Same in k1 and tsl.
    "n_genwook_grts1",
    "n_genwook_coms1",
    "n_genwook_coms2",
    "n_genwook_comm1",
    "n_genwook_comm2",
    "n_genwook_angs",
    "n_genwook_angm",
    "n_genwook_ques",
    "n_genwook_quem",
    "n_genwook_scrs",
    "n_genwook_scrm",
    "n_genwook_lghs",
    "n_genwook_lghm",
    "n_genwook_hpys",
    "n_genwook_hpym",
    "n_genwook_sads",
    "n_genwook_sadm",
    "n_genfwook_grts1",
    "n_genfwook_coms1",
    "n_genfwook_coms2",
    "n_genfwook_comm1",
    "n_genfwook_comm2",
    "n_genfwook_angs",
    "n_genfwook_angm",
    "n_genfwook_ques",
    "n_genfwook_quem",
    "n_genfwook_scrs",
    "n_genfwook_scrm",
    "n_genfwook_lghs",
    "n_genfwook_lghm",
    "n_genfwook_hpys",
    "n_genfwook_hpym",
    "n_genfwook_sads",
    "n_genfwook_sadm",
    "n_gentwook_grts1",
    "n_gentwook_coms1",
    "n_gentwook_coms2",
    "n_gentwook_comm1",
    "n_gentwook_comm2",
    "n_gentwook_angs",
    "n_gentwook_angm",
    "n_gentwook_ques",
    "n_gentwook_quem",
    "n_gentwook_scrs",
    "n_gentwook_scrm",
    "n_gentwook_lghs",
    "n_gentwook_lghm",
    "n_gentwook_hpys",
    "n_gentwook_hpym",
    "n_gentwook_sads",
    "n_gentwook_sadm",
    "n_genwwook_grts1",
    "n_genwwook_coms1",
    "n_genwwook_coms2",
    "n_genwwook_comm1",
    "n_genwwook_comm2",
    "n_genwwook_angs",
    "n_genwwook_angm",
    "n_genwwook_ques",
    "n_genwwook_quem",
    "n_genwwook_scrs",
    "n_genwwook_scrm",
    "n_genwwook_lghs",
    "n_genwwook_lghm",
    "n_genwwook_hpys",
    "n_genwwook_hpym",
    "n_genwwook_sads",
    "n_genwwook_sadm",
    "n_gftwilek_grts",
    "n_gftwilek_grtm",
    "n_gftwilek_coms1",
    "n_gftwilek_coms2",
    "n_gftwilek_comm1",
    "n_gftwilek_comm2",
    "n_gftwilek_coml1",
    "n_gftwilek_coml2",
    "n_gftwilek_angs",
    "n_gftwilek_angm",
    "n_gftwilek_angl",
    "n_gftwilek_ques",
    "n_gftwilek_quel",
    "n_gftwilek_scrs",
    "n_gftwilek_scrm",
    "n_gftwilek_scrl",
    "n_gftwilek_ples",
    "n_gftwilek_plem",
    "n_gftwilek_lghs",
    "n_gftwilek_lghm",
    "n_gftwilek_hpys",
    "n_gftwilek_hpym",
    "n_gftwilek_hpyl",
    "n_gftwilek_sads",
    "n_gftwilek_sadm",
    "n_gftwilek_sadl",
    "n_gftwilek_seds",
    "n_gftwilek_sedm",
    "n_gftwilek_sedl",
    "n_gmtwilek_grts",
    "n_gmtwilek_grtm",
    "n_gmtwilek_coms1",
    "n_gmtwilek_coms2",
    "n_gmtwilek_comm1",
    "n_gmtwilek_comm2",
    "n_gmtwilek_coml1",
    "n_gmtwilek_coml2",
    "n_gmtwilek_angs",
    "n_gmtwilek_angm",
    "n_gmtwilek_angl",
    "n_gmtwilek_ques",
    "n_gmtwilek_quem",
    "n_gmtwilek_quel",
    "n_gmtwilek_scrs",
    "n_gmtwilek_scrm",
    "n_gmtwilek_scrl",
    "n_gmtwilek_ples",
    "n_gmtwilek_plem",
    "n_gmtwilek_lghs",
    "n_gmtwilek_lghm",
    "n_gmtwilek_hpys",
    "n_gmtwilek_hpym",
    "n_gmtwilek_hpyl",
    "n_gmtwilek_sads",
    "n_gmtwilek_sadm",
    "n_gmtwilek_sadl",
    "n_gmtwilek_seds",
    "n_gmtwilek_sedm",
    "n_gmtwilek_sedl",
    "n_gtmtwlek_grts",
    "n_gtmtwlek_grtm",
    "n_gtmtwlek_coms1",
    "n_gtmtwlek_coms2",
    "n_gtmtwlek_comm1",
    "n_gtmtwlek_comm2",
    "n_gtmtwlek_coml1",
    "n_gtmtwlek_coml2",
    "n_gtmtwlek_angs",
    "n_gtmtwlek_angm",
    "n_gtmtwlek_angl",
    "n_gtmtwlek_ques",
    "n_gtmtwlek_quem",
    "n_gtmtwlek_quel",
    "n_gtmtwlek_scrs",
    "n_gtmtwlek_scrm",
    "n_gtmtwlek_scrl",
    "n_gtmtwlek_ples",
    "n_gtmtwlek_plem",
    "n_gtmtwlek_lghs",
    "n_gtmtwlek_lghm",
    "n_gtmtwlek_hpys",
    "n_gtmtwlek_hpym",
    "n_gtmtwlek_hpyl",
    "n_gtmtwlek_sads",
    "n_gtmtwlek_sadm",
    "n_gtmtwlek_sadl",
    "n_grodian_grts",
    "n_grodian_grtm",
    "n_grodian_coms1",
    "n_grodian_coms2",
    "n_grodian_comm1",
    "n_grodian_comm2",
    "n_grodian_coml1",
    "n_grodian_coml2",
    "n_grodian_angs",
    "n_grodian_angm",
    "n_grodian_angl",
    "n_grodian_ques",
    "n_grodian_quem",
    "n_grodian_quel",
    "n_grodian_scrs",
    "n_grodian_scrm",
    "n_grodian_scrl",
    "n_grodian_ples",
    "n_grodian_plem",
    "n_grodian_lghs",
    "n_grodian_lghm",
    "n_grodian_hpys",
    "n_grodian_hpym",
    "n_grodian_hpyl",
    "n_grodian_sads",
    "n_grodian_sadm",
    "n_grodian_sadl",
    "n_gtrodian_grts",
    "n_gtrodian_grtm",
    "n_gtrodian_coms1",
    "n_gtrodian_coms2",
    "n_gtrodian_comm1",
    "n_gtrodian_comm2",
    "n_gtrodian_coml1",
    "n_gtrodian_coml2",
    "n_gtrodian_angs",
    "n_gtrodian_angm",
    "n_gtrodian_angl",
    "n_gtrodian_ques",
    "n_gtrodian_quem",
    "n_gtrodian_quel",
    "n_gtrodian_scrs",
    "n_gtrodian_scrm",
    "n_gtrodian_scrl",
    "n_gtrodian_ples",
    "n_gtrodian_plem",
    "n_gtrodian_lghs",
    "n_gtrodian_lghm",
    "n_gtrodian_hpys",
    "n_gtrodian_hpym",
    "n_gtrodian_hpyl",
    "n_gtrodian_sads",
    "n_gtrodian_sadm",
    "n_gtrodian_sadl",
    "n_grakata_grts",
    "n_grakata_grtm",
    "n_grakata_coms1",
    "n_grakata_coms2",
    "n_grakata_comm1",
    "n_grakata_comm2",
    "n_grakata_coml1",
    "n_grakata_coml2",
    "n_grakata_angs",
    "n_grakata_angm",
    "n_grakata_angl",
    "n_grakata_ques",
    "n_grakata_quem",
    "n_grakata_quel",
    "n_grakata_scrs",
    "n_grakata_scrm",
    "n_grakata_scrl",
    "n_grakata_ples",
    "n_grakata_plem",
    "n_grakata_lghs",
    "n_grakata_lghm",
    "n_grakata_hpys",
    "n_grakata_hpym",
    "n_grakata_hpyl",
    "n_grakata_sads",
    "n_grakata_sadm",
    "n_grakata_sadl",
    "n_genwrakata_grts",
    "n_genwrakata_grtm",
    "n_genwrakata_coms1",
    "n_genwrakata_coms2",
    "n_genwrakata_comm1",
    "n_genwrakata_comm2",
    "n_genwrakata_coml1",
    "n_genwrakata_coml2",
    "n_genwrakata_angs",
    "n_genwrakata_angm",
    "n_genwrakata_angl",
    "n_genwrakata_ques",
    "n_genwrakata_quem",
    "n_genwrakata_quel",
    "n_genwrakata_scrs",
    "n_genwrakata_scrm",
    "n_genwrakata_scrl",
    "n_genwrakata_ples",
    "n_genwrakata_plem",
    "n_genwrakata_lghs",
    "n_genwrakata_lghm",
    "n_genwrakata_hpys",
    "n_genwrakata_hpym",
    "n_genwrakata_hpyl",
    "n_genwrakata_sads",
    "n_genwrakata_sadm",
    "n_genwrakata_sadl",
    "n_genjawa_grts1",
    "n_genjawa_grts2",
    "n_genjawa_coms1",
    "n_genjawa_coms2",
    "n_genjawa_comm1",
    "n_genjawa_comm2",
    "n_genjawa_ques",
    "n_genjawa_quem",
    "n_genhutt_grts",
    "n_genhutt_coms1",
    "n_genhutt_coms2",
    "n_genhutt_comm1",
    "n_genhutt_comm2",
    "n_genhutt_coml1",
    "n_genhutt_coml2",
    "n_genhutt_angm",
    "n_genhutt_angl",
    "n_genhutt_ques",
    "n_genhutt_quem",
    "n_genhutt_quel",
    "n_genhutt_scrs",
    "n_genhutt_scrm",
    "n_genhutt_lghs",
    "n_genhutt_lghm",
    "n_genhutt_hpys",
    "n_genhutt_hpyl",
    "n_genduros_grt",
    "n_genduros_coms1",
    "n_genduros_coms2",
    "n_genduros_comm1",
    "n_genduros_comm2",
    "n_genduros_coml1",
    "n_genduros_coml2",
    "n_genbith_grt",
    "n_genbith_coms1",
    "n_genbith_comm1",
    "n_genbith_coml1",
}
